// serial.h

#ifndef _SERIAL_
#define _SERIAL_

//---------------------------------------------------------------------------

#include <cfifo.h>

//---------------------------------------------------------------------------

#define SERIAL_COM1  				0x00
#define SERIAL_COM2  				0x01
#define SERIAL_INVALIDPORT			0xFFFF

#define SERIAL_BIT5   				0x00
#define SERIAL_BIT6   				0x01
#define SERIAL_BIT7   				0x02
#define SERIAL_BIT8	     			0x03

#define SERIAL_STOP1     			0x00
#define SERIAL_STOP2     			0x04

#define SERIAL_PARITYNOT   		0x00
#define SERIAL_PARITYODD   		0x08
#define SERIAL_PARITYEVEN  		0x18

#define SERIAL_BAUD1200     		96
#define SERIAL_BAUD2400      		48
#define SERIAL_BAUD4800				24
#define SERIAL_BAUD9600      		12
#define SERIAL_BAUD19200	     	6

#define SERIAL_E_OK					0
#define SERIAL_E_INIT				1

#define SERIAL_RBUFFSTATE			1
#define SERIAL_WBUFFSTATE			2
#define SERIAL_ERRORSTATE			4

//---------------------------------------------------------------------------

#ifdef __cplusplus
	 #define __CPPARGS ...
#else
	 #define __CPPARGS
#endif

//---------------------------------------------------------------------------
typedef struct
{
 unsigned short		rw_size;
 unsigned short		speed;
 unsigned char			bit;
 unsigned char			stopbit;
 unsigned char			parity;
}serial_TInit;
//---------------------------------------------------------------------------
unsigned short serial_Init(unsigned short portid,serial_TInit *cfg);
unsigned short serial_Init(unsigned short portid,
									unsigned char 	speed,
									unsigned char 	bit,
									unsigned char 	stopbit,
									unsigned char 	parity,
									unsigned short rw_size);

void 				serial_Destroy(unsigned short portid);
void				serial_Reset(unsigned short portid);
void				serial_ResetR(unsigned short portid);
void				serial_ResetW(unsigned short portid);

unsigned short serial_Read(unsigned short portid,unsigned char *data,unsigned short size);
unsigned short serial_Write(unsigned short portid,unsigned char *data,unsigned short size);

void 				serial_SetRTS(unsigned short portid);
void 				serial_SetDTR(unsigned short portid);
void 				serial_ResetRTS(unsigned short portid);
void 				serial_ResetDTR(unsigned short portid);

unsigned char 	serial_ReadDCD(unsigned short portid);
unsigned char 	serial_ReadCTS(unsigned short nport);
unsigned char 	serial_ReadRI(unsigned short portid);
unsigned char 	serial_ReadDSR(unsigned short portid);

unsigned short serial_GetWBSize(unsigned short portid);
unsigned short serial_GetRBSize(unsigned short portid);
unsigned short serial_GetWBFree(unsigned short portid);
unsigned short serial_GetRBFree(unsigned short portid);

//---------------------------------------------------------------------------

#endif