// serial.cpp

#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <mem.h>
#include <timer.h>
#include <cfifo.h>
#include <serial.h>
//---------------------------------------------------------------------------------
#define RW_TEST						10
#define RW_ENDTEST					0
#define RW_DISABLE					0
#define RW_READ						1
#define RW_WRITE						2
#define RW_READWRITE					3
//---------------------------------------------------------------------------------
#define SREG_RX		0
#define SREG_TX		0
#define SREG_IE		1
#define SREG_II		2
#define SREG_LC		3
#define SREG_MC		4
#define SREG_LS		5
#define SREG_MS		6
#define SREG_DL		0
#define SREG_DH		1
//---------------------------------------------------------------------------------
#define SINTR_PORT_0		0x0C
#define SINTR_PORT_1		0x0B
#define SINTR_IMR			0x21
#define SINTR_ICR			0x20
#define INTR_READ			1
#define INTR_WRITE		2
//---------------------------------------------------------------------------------
#define SRMS_ALL		0xFF
#define SRMS_CTS_CHANGE		0x01
#define SRMS_SR_CHANGE		0x02
#define SRMS_RI_CHANGE		0x04
#define SRMS_DC_CHANGE		0x08
#define SRMS_CTS				0x10
#define SRMS_DSR				0x20
#define SRMS_RI				0x40
#define SRMS_DCD				0x80
//---------------------------------------------------------------------------------
#define SRMC_ALL		0xFF
#define SRMC_DTR		0x01
#define SRMC_RTS		0x02
#define SRMC_OUT1		0x04
#define SRMC_OUT2		0x08
#define SRMC_RING		0x10
//---------------------------------------------------------------------------------
#ifdef __cplusplus
	 #define __CPPARGS ...
#else
	 #define __CPPARGS
#endif
//---------------------------------------------------------------------------------
typedef struct
{
 unsigned char  	port;
 unsigned short 	speed;
 unsigned char  	cfg;
 unsigned char  	fReadWrite;

 unsigned char 	fInit;
 unsigned char 	fISR;
 unsigned short 	baseadress;
 unsigned short 	intradress;

 unsigned char  	oldcfg;
 unsigned char  	oldie;
 unsigned char  	oldmc;
 unsigned char  	oldspeed;
 unsigned char  	oldisr;
 void _interrupt _far (*old_int)(__CPPARGS);
 unsigned char 	lock;

 CFifo 		*rbuffer;
 CFifo	 	*wbuffer;
}TPort;
//---------------------------------------------------------------------------------
static TPort Port[4];
static unsigned long error_code;
//---------------------------------------------------------------------------------
void serial_Error(unsigned short code)
{
 error_code=code;
}
//---------------------------------------------------------------------------------
void serial_EnableISR(TPort *port,unsigned char flag)
{
 unsigned char ie;
 if (port->fReadWrite==RW_TEST)
 {
  outp(port->baseadress+SREG_IE,2);
 }
 else
 {
  ie=inp(port->baseadress+SREG_IE);
  outp(port->baseadress+SREG_IE,ie|flag);
 }
}
//---------------------------------------------------------------------------------
void serial_DisableISR(TPort *port,unsigned char flag)
{
 unsigned char ie=inp(port->baseadress+SREG_IE);
 outp(port->baseadress+SREG_IE,(ie&(~flag))&3);
 if (port->fReadWrite==RW_TEST) port->fReadWrite=RW_ENDTEST;
}
//---------------------------------------------------------------------------------
void far process_ReadIRQ(TPort *port)
{
 char byte=inp(port->baseadress);
// cprintf("%02X",byte);
 port->rbuffer->Write(&byte,1);
}
//---------------------------------------------------------------------------------
void far process_WriteIRQ(TPort *port)
{
 unsigned char byte;
 port->wbuffer->Read(&byte,1);
 outp(port->baseadress,byte);
 if (!port->wbuffer->GetSize())
 {
	serial_DisableISR(port,RW_WRITE);
 }
}
//---------------------------------------------------------------------------------
void far process_IRQ(TPort *port)
{
 if (port->fReadWrite==RW_TEST)
 {
  outp(port->baseadress,0);
  serial_DisableISR(port,RW_WRITE);
 }
 else if (port->fReadWrite==RW_READ) process_ReadIRQ(port);
 else if (port->fReadWrite==RW_WRITE) process_WriteIRQ(port);
 else if (port->fReadWrite==RW_READWRITE)
 {
  do
  {
	if (inportb(port->baseadress+SREG_II)&4) process_ReadIRQ(port);
	else process_WriteIRQ(port);
  }
  while (!(inp(port->baseadress+SREG_II)&1));
 }
}
//---------------------------------------------------------------------------------
void _interrupt _far new_IRQ4(__CPPARGS)
{
 Port[0].lock=1;
 process_IRQ(&Port[0]);
 Port[0].lock=0;
 outp(0x20,0x20);
}
//---------------------------------------------------------------------------------
void _interrupt _far new_IRQ3(__CPPARGS)
{
 Port[1].lock=1;
 process_IRQ(&Port[1]);
 Port[1].lock=0;
 outp(0x20,0x20);
}
//---------------------------------------------------------------------------------
unsigned short serial_GetAdress(unsigned short port)
{
 port*=2;
 return *(unsigned short*)MK_FP(0x40,port);
}
//---------------------------------------------------------------------------------
unsigned short serial_GetIntrAdress(unsigned short port)
{
 if (port==SERIAL_COM1) return SINTR_PORT_0;
 else return SINTR_PORT_1;
}
//---------------------------------------------------------------------------------
void serial_SetPortSpeed(unsigned short adress,unsigned short speed)
{
 outp(adress+SREG_LC,0x80);
 outp(adress+SREG_DL,(unsigned char)(speed&0xFF));
 outp(adress+SREG_DH,(unsigned char)(speed>>8));
}
//---------------------------------------------------------------------------------
unsigned short serial_GetPortSpeed(unsigned short adress)
{
 unsigned short speed;
 unsigned char lc;
 asm cli;
 lc=inp(adress+SREG_LC);
 outp(adress+SREG_LC,0x80);
 speed=inp(adress+SREG_DL);
 speed+=((unsigned short)inp(adress+SREG_DH))<<8;
 outp(adress+SREG_LC,lc);
 asm sti;
 return speed;
}
//---------------------------------------------------------------------------------
void serial_StoreOldParam(TPort *port)
{
 asm cli;
 port->oldie=inp(port->baseadress+SREG_IE);
 port->oldmc=inp(port->baseadress+SREG_MC);
 port->oldcfg=inp(port->baseadress+SREG_LC);
 port->oldspeed=serial_GetPortSpeed(port->baseadress);
 if (port->port==SERIAL_COM1)
 {
  port->oldisr=inp(0x21)&16;
 }
 else port->oldisr=inp(0x21)&8;
 asm sti;
}
//---------------------------------------------------------------------------------
void serial_RestoreOldParam(TPort *port)
{
 unsigned char isr;
 while (port->lock);
 asm cli;
 outp(port->baseadress+SREG_IE,0);
 outp(port->baseadress+SREG_MC,0);
 isr=inp(0x21);
 isr|=port->oldisr;
 outp(0x21,isr);
 asm sti;
 _dos_setvect(port->intradress,port->old_int);
 outp(port->baseadress+SREG_IE,port->oldie);
 serial_SetPortSpeed(port->baseadress,port->oldspeed);
 outp(port->baseadress+SREG_LC,port->oldcfg);
 outp(port->baseadress+SREG_MC,port->oldmc);
}
//---------------------------------------------------------------------------------
void serial_SetNewParam(TPort *port)
{
 unsigned char isr;
 asm cli;
 outp(port->baseadress+SREG_IE,0);
 outp(port->baseadress+SREG_MC,0);
 asm sti
 serial_SetPortSpeed(port->baseadress,port->speed);
 isr=inp(0x21);
 outp(port->baseadress+SREG_LC,port->cfg);

 port->old_int=_dos_getvect(port->intradress);
 if (port->port==SERIAL_COM1) _dos_setvect(port->intradress,new_IRQ4);
 else _dos_setvect(port->intradress,new_IRQ3);

 asm cli;
 outp(0x21,(port->port==SERIAL_COM1)?isr&0xEF:isr&0xF7);
 asm sti;
 outp(port->baseadress+SREG_MC,0x08);
}
//---------------------------------------------------------------------------------
unsigned short serial_Init(unsigned short portid,
									unsigned char 	speed,
									unsigned char 	bit,
									unsigned char 	stopbit,
									unsigned char 	parity,
									unsigned short rw_size)

{
 unsigned long btime;
 TPort *port=&Port[portid];
 if (port->fInit) return SERIAL_INVALIDPORT;
 port->port=portid;
 port->cfg=bit|stopbit|parity;
 port->speed=speed;
 port->rbuffer=new CFifo(rw_size);
 port->wbuffer=new CFifo(rw_size);
 port->baseadress=serial_GetAdress(portid);
 port->intradress=serial_GetIntrAdress(portid);
 serial_StoreOldParam(port);
 serial_SetNewParam(port);
 port->fInit=1;
 btime=Timer.Clock();
 port->fReadWrite=RW_TEST;
 serial_EnableISR(port,RW_READ);

 while (port->fReadWrite==RW_TEST)
 {
  if (Timer.Interval(btime)>1000)
  {
	serial_Destroy(portid);
	return SERIAL_INVALIDPORT;
  }
 }
 port->fReadWrite=RW_READWRITE;
 serial_EnableISR(port,RW_READ);
 return port->port;
}
//---------------------------------------------------------------------------------
void serial_Destroy(unsigned short portid)
{
 TPort *port=&Port[portid];

 if (port->fInit)
 {
  serial_RestoreOldParam(port);
  if (port->rbuffer!=NULL) delete port->rbuffer;
  port->rbuffer=NULL;
  if (port->wbuffer!=NULL) delete port->wbuffer;
  port->wbuffer=NULL;

  port->fInit=0;
 }
}
//---------------------------------------------------------------------------------
unsigned short serial_Read(unsigned short nport,unsigned char *data,unsigned short size)
{
 TPort *port=&Port[nport];
 if (port->rbuffer!=NULL)
 {
  while (port->lock);
  return port->rbuffer->Read(data,size);
 }
 else return 0;
}
//---------------------------------------------------------------------------------
unsigned short serial_Write(unsigned short nport,unsigned char *data,unsigned short size)
{
 TPort *port=&Port[nport];
 unsigned short result=0;

 if (port->wbuffer!=NULL)
 {
  while (port->lock);
  result=port->wbuffer->Write(data,size);
  if (port->wbuffer->GetSize()) serial_EnableISR(port,RW_WRITE);
 }
 return result;
}
//---------------------------------------------------------------------------------
void serial_SetRTS(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MC);
 outportb(port->baseadress+SREG_MC,SRMC_RTS | rvalue);
}
//---------------------------------------------------------------------------------
void serial_ResetRTS(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MC);
 outportb(port->baseadress+SREG_MC,(~SRMC_RTS) & rvalue);
}
//---------------------------------------------------------------------------------
void serial_SetDTR(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MC);
 outportb(port->baseadress+SREG_MC,SRMC_DTR | rvalue);
}
//---------------------------------------------------------------------------------
void serial_ResetDTR(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MC);
 outportb(port->baseadress+SREG_MC,(~SRMC_DTR) & rvalue);
}
//---------------------------------------------------------------------------------
unsigned char serial_ReadDCD(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MS);
 if (rvalue&SRMS_DCD) return 1;
 else return 0;
}
//---------------------------------------------------------------------------------
unsigned char serial_ReadRI(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MS);
 if (rvalue&SRMS_RI) return 1;
 else return 0;
}
//---------------------------------------------------------------------------------
unsigned char serial_ReadDSR(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MS);
 if (rvalue&SRMS_DSR) return 1;
 else return 0;
}
//---------------------------------------------------------------------------------
unsigned char serial_ReadCTS(unsigned short nport)
{
 TPort *port=&Port[nport];
 unsigned char rvalue=inportb(port->baseadress+SREG_MS);
 if (rvalue&SRMS_CTS) return 1;
 else return 0;
}
//---------------------------------------------------------------------------------
unsigned short  serial_GetWBSize(unsigned short portid)
{
 TPort *port=&Port[portid];
 return port->wbuffer->GetSize();
}
//---------------------------------------------------------------------------------
unsigned short  serial_GetRBSize(unsigned short portid)
{
 TPort *port=&Port[portid];
 return port->rbuffer->GetSize();
}
//---------------------------------------------------------------------------------
unsigned short  serial_GetWBFree(unsigned short portid)
{
 TPort *port=&Port[portid];
 return port->wbuffer->GetFree();
}
//---------------------------------------------------------------------------------
unsigned short  serial_GetRBFree(unsigned short portid)
{
 TPort *port=&Port[portid];
 return port->rbuffer->GetFree();
}
//---------------------------------------------------------------------------------
void serial_Reset(unsigned short portid)
{
 TPort *port=&Port[portid];
 while (port->lock);
 port->rbuffer->Reset();
 port->wbuffer->Reset();
}
//---------------------------------------------------------------------------------
void serial_ResetR(unsigned short portid)
{
 TPort *port=&Port[portid];
 port->rbuffer->Reset();
}
//---------------------------------------------------------------------------------
void serial_ResetW(unsigned short portid)
{
 TPort *port=&Port[portid];
 port->wbuffer->Reset();
}

unsigned short serial_Init(unsigned short portid,serial_TInit *cfg)
{
   return serial_Init(portid,cfg->speed,cfg->bit,cfg->stopbit,cfg->parity,cfg->rw_size);
}
